"""Progress model class."""

import random

from .quiz import Quiz, Quizzes
from .retention import Retention
from .topic import Topics


class Progress:
    """Keep track of progress on quizzes."""
    def __init__(self, progress_dict: dict[str, dict[str, str | int]]) -> None:
        self.__progress_dict = {key: Retention.from_dict(value) for key, value in progress_dict.items()}
        self.__current_quiz: Quiz | None = None

    def update(self, quiz: Quiz, correct: bool) -> None:
        """Update the progress on the quiz."""
        self.__progress_dict.setdefault(str(quiz), Retention()).update(correct)

    def get_retention(self, quiz: Quiz) -> Retention:
        """Return the quiz retention."""
        return self.__progress_dict.get(str(quiz), Retention())

    def next_quiz(self, topics: Topics) -> Quiz | None:
        """Return the next quiz."""
        for must_have_progress in (True, False):
            for topic in topics:
                if quizzes := self.__eligible_quizzes(topic.quizzes, must_have_progress):
                    self.__current_quiz = random.choice(list(quizzes))
                    return self.__current_quiz
        self.__current_quiz = None
        return None

    def __eligible_quizzes(self, quizzes: Quizzes, must_have_progress: bool) -> Quizzes:
        """Return the quizzes that can be the next quiz."""
        eligible = [quiz for quiz in quizzes if not self.__is_silenced(quiz) and quiz != self.__current_quiz]
        eligible_with_progress = [quiz for quiz in eligible if self.__has_progress(quiz)]
        return set(eligible_with_progress if must_have_progress else eligible_with_progress or eligible)

    def __is_silenced(self, quiz: Quiz) -> bool:
        """Is the quiz silenced?"""
        return self.get_retention(quiz).is_silenced()

    def __has_progress(self, quiz: Quiz) -> bool:
        """Has the quiz been presented to the user before?"""
        return str(quiz) in self.__progress_dict

    def as_dict(self) -> dict[str, dict[str, int | str]]:
        """Return the progress as dict."""
        return {key: value.as_dict() for key, value in self.__progress_dict.items()}



"""Store and load progress data."""

import pathlib

from ..metadata import NAME
from ..model import Progress
from ..ui.text import show_error_and_exit

from .json_file import load_json, dump_json


PROGRESS_JSON = pathlib.Path.home() / f".{NAME.lower()}-progress.json"


def load_progress() -> Progress:
    """Load the progress from the user's home folder."""
    try:
        progress_dict = load_json(PROGRESS_JSON, default={})
        return Progress(progress_dict)
    except Exception as reason:  # pylint: disable=broad-except
        show_error_and_exit(
            f"""{NAME} cannot parse the progress information in {PROGRESS_JSON}: {reason}.
To fix this, remove or rename {PROGRESS_JSON} and start {NAME} again. Unfortunately, this will reset your progress.
Please consider opening a bug report at https://github.com/fniessink/{NAME.lower()}. Be sure to attach the invalid
progress file to the issue.
""")


def save_progress(progress: Progress) -> None:
    """Save the progress to the user's home folder."""
    dump_json(PROGRESS_JSON, progress.as_dict())



"""Command to show progress information."""

from datetime import datetime

from rich.table import Table

from toisto.metadata import Language, SUPPORTED_LANGUAGES
from toisto.model import Progress, Topics
from toisto.ui.text import console, format_duration, format_datetime


def show_progress(language: Language, topics: Topics, progress: Progress) -> None:
    """Show progress."""
    table = Table(title=f"Progress {SUPPORTED_LANGUAGES[language]}")
    table.add_column("Quiz type")
    table.add_column("Question")
    table.add_column("From")
    table.add_column("To")
    table.add_column("Answer(s)")
    table.add_column("Attempts", justify="right")
    table.add_column("Retention")
    table.add_column("Not quizzed until")
    sorted_quizzes = sorted(topics.quizzes, key=lambda quiz: progress.get_retention(quiz).count, reverse=True)
    for quiz in sorted_quizzes:
        retention = progress.get_retention(quiz)
        skip = retention.skip_until
        table.add_row(
            quiz.quiz_type.capitalize(),
            quiz.question,
            quiz.question_language,
            quiz.answer_language,
            "\n".join(quiz.answers),
            str(retention.count),
            format_duration(retention.length) if retention.length else "",
            format_datetime(skip) if skip and skip > datetime.now() else ""
        )
    with console.pager():
        console.print(table)



"""Main module for the application."""

import readline  # pylint: disable=unused-import

from .ui.cli import parser
from .command import practice, show_progress
from .persistence import load_quizzes, load_progress


def main():
    """Main program."""
    args = parser.parse_args()
    topics = load_quizzes(args.language, args.source_language, args.topic, args.topic_file)
    progress = load_progress()
    if args.command == "practice":
        practice(topics, progress)
    else:
        show_progress(args.language, topics, progress)



"""Load quizzes from topic files."""

import pathlib

from ..metadata import NAME, TOPIC_JSON_FILES, Language
from ..model import concept_factory, Topic, Topics
from ..ui.text import show_error_and_exit

from .json_file import load_json


def load_quizzes(
    language: Language, source_language: Language, builtin_topics_to_load: list[str], topic_files_to_load: list[str]
) -> Topics:
    """Load the concepts from the topics and generate the quizzes."""
    topics = set()
    topic_files: list[pathlib.Path] = []
    if builtin_topics_to_load or topic_files_to_load:
        topic_files.extend(topic for topic in TOPIC_JSON_FILES if topic.stem in builtin_topics_to_load)
        topic_files.extend(pathlib.Path(topic) for topic in topic_files_to_load)
    else:
        topic_files.extend(TOPIC_JSON_FILES)
    for topic_file in topic_files:
        topic_quizzes = set()
        try:
            for concept_dict in load_json(topic_file):
                concept = concept_factory(concept_dict)
                topic_quizzes.update(concept.quizzes(language, source_language))
        except Exception as reason:  # pylint: disable=broad-except
            show_error_and_exit(f"{NAME} cannot read topic {topic_file}: {reason}.\n")
        topics.add(Topic(topic_file.stem, topic_quizzes))
    return Topics(topics)



"""Output for the user."""

from datetime import datetime, timedelta
from typing import NoReturn
import sys

from rich.console import Console
from rich.theme import Theme

from .diff import colored_diff
from ..metadata import NAME, VERSION
from ..model import Label, Quiz, Retention


theme = Theme({
    "secondary": "grey69",
    "quiz": "medium_purple1",
    "inserted": "bright_green",
    "deleted": "bright_red"
})

console = Console(theme=theme)

WELCOME = f"""👋 Welcome to {NAME} [white not bold]v{VERSION}[/white not bold]!

Practice as many words and phrases as you like, for as long as you like.
Hit Ctrl-C or Ctrl-D to quit.

[secondary]{NAME} quizzes you on words and phrases repeatedly. Each time you answer a
quiz correctly, {NAME} will wait longer before repeating it. If you answer
a quiz incorrectly, you get one additional attempt to give the correct
answer. If the second attempt is not correct either, {NAME} will reset the
quiz interval.[/secondary]
"""

DONE = f"""👍 Good job. You're done for now. Please come back later or try a different topic.
[secondary]Type `{NAME.lower()} -h` for more information.[/secondary]
"""

TRY_AGAIN = "⚠️  Incorrect. Please try again."

CORRECT = "✅ Correct.\n"


def format_duration(duration: timedelta) -> str:
    """Format the duration in a human friendly way."""
    if duration.days > 1:
        return f"{duration.days} days"
    seconds = duration.seconds + (24 * 3600 if duration.days else 0)
    if seconds >= 1.5 * 3600:  # 1.5 hours
        hours = round(seconds / 3600)
        return f"{hours} hours"
    if seconds >= 1.5 * 60:  # 1.5 minutes
        minutes = round(seconds / 60)
        return f"{minutes} minutes"
    return f"{seconds} seconds"


def format_datetime(date_time: datetime) -> str:
    """Return a human readable version of the datetime"""
    return date_time.isoformat(sep=" ", timespec="minutes")


def feedback_correct(guess: Label, quiz: Quiz, quiz_progress: Retention) -> str:
    """Return the feedback about a correct result."""
    text = CORRECT
    if quiz_progress.skip_until:
        silence_duration = format_duration(quiz_progress.skip_until - datetime.now())
        long_initial_skip_explanation = " as you already seem to know this" if quiz_progress.count == 1 else ""
        text += f"[secondary]Skipping this quiz for {silence_duration}{long_initial_skip_explanation}.[/secondary]\n"
    if other_answers := quiz.other_answers(guess):
        label = "Another correct answer is" if len(other_answers) == 1 else "Other correct answers are"
        enumerated_answers = ", ".join([f'"{answer}"' for answer in other_answers])
        text += f"""[secondary]{label} {enumerated_answers}.[/secondary]\n"""
    return text


def feedback_incorrect(guess: Label, quiz: Quiz) -> str:
    """Return the feedback about an incorrect result."""
    diff = colored_diff(guess, quiz.answer)
    return f'❌ Incorrect. The correct answer is "{diff}".\n'


def instruction(quiz: Quiz) -> str:
    """Return the instruction for the quiz."""
    return f"[quiz]{quiz.instruction()}:[/quiz]"


def show_error_and_exit(message: str) -> NoReturn:
    """Print the error message to stderr and exit."""
    sys.stderr.write(message)
    sys.exit(2)
